#include "config.h"
extern "C" {
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
}

#include "globals.hh"
#include "gui.hh"
#include "loader.hh"
#include "serial.hh"
#include "strings.hh"

const string Loader::bootString[] = {
    "S12300008E00FF4FB71035CE10001F2E80FC1D282086B0A72B860CA72D8D2B20FC7F00E84F",
    "S12300202004860197E88D5F167D00E827048D102006183C8D5F183808180926E939188F95",
    "S1230040E700A600202D863E8D298D3B817327F6368D2C8F8D29188F32817227E58170274B",
    "S123006032815727BD815027B48142273F817727CD863F3CCE10001F2E80FCA72F38398D51",
    "S123008006368D031632393CCE10001F2E20FCA62F3839188FE10027A9A60043D7E994E9C8",
    "S12300A0270486168D2386028D1F20968606B7103BB7103F8A01B7103B183C9DD918387F21",
    "S12300C0103B188F8602CE103FB7103BE7008A01B7103B8D0A7F103B3918CE6F9A2004184A",
    "S10B00E0CE0B29180926FC3996",
    "S9030000FC"
};
  
Loader::Loader() {
	progStream = NULL;

	options.put(*(new Option(str.BLKERASELOPT, str.BLKERASESOPT, POPT_ARG_LONG,   &bulkErase,         0, str.BLKERASETIP, string(""))));
	options.put(*(new Option(str.BTSTRPLOPT,   str.BTSTRPSOPT,   POPT_ARG_LONG,   &bootstrapFeedback, 0, str.BTSTRPTIP,   string(""))));
	options.put(*(new Option(str.HDWRLOPT,     str.HDWRSOPT,     POPT_ARG_LONG,   &hardwareEcho,      0, str.HDWRTIP,     string(""))));
	options.put(*(new Option(str.MNEELOPT,     str.MNEESOPT,     POPT_ARG_STRING, &minEEPROM,         0, str.MNEETIP,     string(""))));
	options.put(*(new Option(str.FILELOPT,     str.FILESOPT,     POPT_ARG_STRING, &fileName,          0, str.FILETIP,     string(""))));
}

// checkPrereqs tests the prerequisites for bringing up the Comm window
// and Download buttons, and sets the program's status appropriately.
// The actual sensitivity control is bundled in setStatus.
void Loader::checkPrereqs() {
	struct stat statbuf;
	int noportname, statfailed, badstat, openfailed, baddev;
	int fd;
	struct termios termios;
	
	if (gui == NULL)
		return;

	noportname = 0;
	statfailed = 0;
	badstat = 0;
	openfailed = 0;
	baddev = 0;

	// make sure we have a port selected, and that the port is a
	// character device, and that we can open the device.  This is
	// weird...  I'm successfully opening the node even though there's
	// no device behind it!  Need to put a termios call in here to
	// check that, even though at this point I'm trespassing pretty
	// heavily on serial.cc's turf...
	if (gui->mainWin.controlPanel.setupPage.getPortName() == "")
		noportname = 1;
	else if ((fd = open(gui->mainWin.controlPanel.setupPage.getPortName().c_str(), O_RDWR)) < 0)
		openfailed = 1;
	else {
		if (fstat(fd, &statbuf) != 0)
			statfailed = 1;
		else if ((statbuf.st_mode & S_IFCHR) == 0)
			badstat = 1;
		else if (tcgetattr(fd, &termios) < 0)
			baddev = 1;
			
		close(fd);
	}

	if ((noportname == 1) || (statfailed == 1) || (badstat == 1) || (openfailed == 1) || (baddev == 1)) {
		gui->setStatus(Gui::NEEDPORT);
		return;
	}
	
	// next question -- do we have a file to download?
	if ((gui->getFileName() == "") ||
		(stat(gui->getFileName().c_str(), &statbuf) != 0) ||
		((statbuf.st_mode & S_IFREG) == 0)) {
		gui->setStatus(Gui::NEEDFILE);
	} else {
		gui->setStatus(Gui::IDLE);
	}
}

bool Loader::openProg() {
	if ((progStream = fopen(gui->getFileName().c_str(), "r")) == NULL) {
		gui->error(str.CANTOPENERR + gui->getFileName(), errno);
		
		return(false);
	}
	
	return(true);
}

bool Loader::closeProg() {
	if (!progStream) {
		gui->error(str.CLOSENULLERR);
		return(false);
	}

	if (fclose(progStream)) {
		gui->error(str.PROGCLOSEERR, errno);
		return(false);
	}

	return(true);
}

int Loader::hexToInt(char ch)
{
	char *buffer;
	
	if ('0' <= ch && ch <= '9') return ch - '0';
	if ('a' <= ch && ch <= 'f') return ch - 'a' + 10;
	if ('A' <= ch && ch <= 'F') return ch - 'A' + 10;

	gui->error(sprintf(gui->errmess, str.BADHEXERR.c_str(), ch));
	return -1;
}

int Loader::decodeHex(const string str, int index)
{
	int hi,lo;
	
	hi = hexToInt(str[index]);
	lo = hexToInt(str[index+1]);
	
	if ((hi < 0) || (lo < 0))
		return(-1);
	
	return (hi * 16) + lo;
}

bool Loader::loadBoot() {
	int i, j;
	gui->setStatus(Gui::BOOTLOAD);

	// init the port at 1200 and download the bootstrap loader
	if (!serial.init1200())
		return(false);

	if (!serial.putChar(0xff, bootstrapFeedback))
		return(false);

	int totalBytes = 0;
	for (j = 0; ; j++) {
		int byteCount;

		if ((bootString[j][0] == 'S') && (bootString[j][1] == '9'))
			break;
		
		byteCount = decodeHex(bootString[j], 2) - 3;
		
		for (i = 0; i < byteCount; i++) {
			int value = decodeHex(bootString[j], (2*i)+8);
			
			if (value < 0)
			  return(-1);
			
			unsigned char dataByte = value;
			
			if (!serial.putChar(dataByte, bootstrapFeedback))
				return(false);
			
			totalBytes++;

			// I have to manually handle the mainloop iterations here
			// or else the gui freezes during a download.  This code
			// used to be in setProgress and setStatus, but that was
			// causing nested calls to setStatus().  The real fix is
			// to thread this app.  Not today.
			gui->setProgress(totalBytes, 256);
			while (gui->events_pending() > 0)
				gui->iteration();
			
			if (gui->getStatus() == Gui::IDLE)
				return (false);

		}

		// Let the port drain so the progress meter is accurate.
		serial.drain();
	}

	// fill up the rest of the first 256 bytes
	while (totalBytes < 256) {
		if (!serial.putChar(0, bootstrapFeedback))
			return(false);
		
		gui->setProgress(totalBytes, 256);
		while (gui->events_pending() > 0)
			gui->iteration();

		if (gui->getStatus() == Gui::IDLE)
			return (false);
		
		totalBytes++;
	}

	return(true);
}

bool Loader::sync() {
	int j;
	
	gui->setStatus(Gui::SYNCWAIT);

	// set the board to 9600 baud
	if (!serial.init9600())
		return(false);
	
	// sync with the board
    serial.putChar('s', false);
    serial.putChar('s', false);
    serial.putChar('s', false);
    serial.putChar('s', false);
    serial.putChar('s', false);

    for (j=0; j<20; j++) {
        int ch, lch;

        // get characters until the last one; return the last valid char
        while (1) {
            lch = serial.getChar();
            if (lch == -1) break;
            ch = lch;
        }

        if (ch == '>') {
                serial.flush();
                break;
        }
		
        serial.putChar('s', false);

        gui->setProgress(j, 20);
		while (gui->events_pending() > 0)
			gui->iteration();
		
		if (gui->getStatus() == Gui::IDLE)
			return (false);
        usleep(55000L);
    }
	
    if (j == 20) {
		gui->error(str.SYNCERR);
        return(false);
    }

	return(true);
}

bool Loader::loadRecord(int startAddr, int numBytes, char line_in[])
{
	gui->dbg(sprintf(gui->errmess, str.RCDDBG.c_str(), startAddr));

	if ((startAddr <= 0xfffe) && ((startAddr + numBytes) > 0xfffe))
		wrotefffe = true;
	
	// use block download to EEPROM mode
	if (!serial.putChar('P', hardwareEcho))
		return(false);
	
	if (!serial.putChar(startAddr >> 8, hardwareEcho))
		return(false);
	if (!serial.putChar(startAddr & 0xff, hardwareEcho))
		return(false);
	if (!serial.putChar(numBytes >> 8, hardwareEcho))		// count hi
		return(false);
	if (!serial.putChar(numBytes & 0xff, hardwareEcho))
		return(false);
	
	// loop sending data bytes from the s-record
	for (int i = 0; i < numBytes; i++) {
		int dataByte = decodeHex(line_in, 2*i);
		if (dataByte < 0) return(false);
		if (!serial.putChar(dataByte, hardwareEcho))
			return(false);
		
		int returnByte = serial.getChar();
		if (dataByte != returnByte) {
			gui->error(sprintf(gui->errmess,
							   str.MEMERR.c_str(),
							   startAddr + i,
							   dataByte & 0xff,
							   returnByte & 0xff));
			return(false);
		}
	}
	
	// board should send prompt now
	if (serial.getChar() != '>') {
		gui->error(str.SYNCERR);
		return(false);
	}
	
	return true;
}

bool Loader::loadProg() {
	char line_in[256];
	struct stat statBuf;

	wrotefffe = false;
	
    if (bulkErase) {
        gui->setStatus(Gui::BULKERASE);
        serial.putChar('B', hardwareEcho);
        serial.putChar(0, hardwareEcho);
        serial.putChar(0, hardwareEcho);
        serial.putChar(0, hardwareEcho);
        serial.putChar(0xF5, hardwareEcho); // config register

        // board should send prompt now
        if (serial.getChar() != '>') {
            gui->error(str.BULKSYNCERR);
			return(false);
        }
    }

	int minEEPROMval = strtol(minEEPROM, NULL, 16);
	
    gui->setStatus(Gui::PROMLOAD);
	// find out how large the file is, so we can work the progress bar
	// correctly
	if (fstat(fileno(progStream), &statBuf)) {
		gui->error(sprintf(gui->errmess, str.STATERR.c_str(), fileName), errno);
		return(false);
	}

	if (statBuf.st_size == 0) {
		gui->error(str.EMPTYERR);
		return(false);
	}

	// loop continues until I've gotten an S9 (termination) line or
	// I've run out of input
	char *rchar;
	do {
		rchar = fgets(line_in,256,progStream);
		
		// Line parsing is a bit ad hoc, but oh well

		// First, lines that aren't S19, S2, or S3 format files
		if ((line_in[0] != 'S') ||
			((line_in[1] > '3') && (line_in[1] < '9'))) {
			gui->error(str.FMTERR);
			return false;
		}

		// If we're looking at an S2 or S3 line the format's OK, but
		// we don't deal with 24 or 32 bit addresses
		if ((line_in[1] == '2') || (line_in[1] == '3')) {
			gui->error(str.S23ERR);
			return false;
		}
			
		// ignore header record (just go to the next line)
		if (line_in[1] == '0') continue;

		// S1 is a data record
		if (line_in[1] == '1') {
			
			int byteCount = decodeHex(line_in, 2) - 3;
			if (byteCount > 253) {
				gui->error(str.BUFTOOSMALLERR);
				return(false);
			}
			
			unsigned int startaddh = decodeHex(line_in, 4);
			unsigned int startaddl = decodeHex(line_in, 6);
			
			if (((startaddh << 8) + startaddl) < minEEPROMval) {
				
				gui->error(sprintf(gui->errmess,
								   str.ADDRERR.c_str(),
								   (startaddh << 8) + startaddl,
								   minEEPROMval));
				return(false);
			}

			if (!loadRecord((startaddh <<  8) + startaddl, byteCount, &line_in[8]))
				return false;

			gui->setProgress(ftell(progStream), statBuf.st_size);
			while (gui->events_pending() > 0)
				gui->iteration();
							 
			if (gui->getStatus() == Gui::IDLE)
				return (false);

		}
	} while ((line_in[1] != '9') && (rchar != NULL));

	// It appears as if the assembler will quite happily construct a
	// program with no start address, letting the user put a start
	// vector in the assembly code "by hand" using FDB at fffe.
	// Worse, at NMSU, we were using a version of the assembler that
	// didn't even support a start address, and I only discovered it
	// when I rebuilt it from sources (why the sources I had didn't
	// match the existing assembler is a mystery to me, too), so we've
	// been telling generations of students to do it that way.  So, I
	// need to apply some sort of heuristic which will handle both S19
	// files without a valid start address, and files which do have
	// one.  I think my most general bet is to see if fffe has already
	// been written during the download; if it has we won't over-write
	// it, while if it has not we'll take a start address from here.
	// This has the additional benefit that if a student forgets to do
	// anything about a start address, their code will end up starting
	// at 0000 which has a very slim chance of working.

	// just for grins, this code allows for the slim chance of no
	// termination record; in that case, it assumes the start address
	// was set by hand.
	if ((line_in[1] == '9') && !wrotefffe) {
		// figure out the starting address
		int starthi = decodeHex(line_in, 4);
		int startlo = decodeHex(line_in, 6);
		
		if (!loadRecord(0xfffe, 2, &line_in[4]))
			return false;
	}

	return true;
}

void Loader::downloadFile() {

	gui->setIO(false);

	// OK, this is a little flaky.  The idea is, I want to execute
	// each of these functions in sequence; if any of them return false, I
	// don't want to execute the rest.  I'm relying on C++'s
	// documented order of evaluation to make this work.  No matter
	// where I bail, I need to clean up properly when I'm done.
	if (serial.openPort() &&
		openProg() &&
		loadBoot() &&
		sync() &&
		loadProg() &&
		closeProg());

	if (gui->winOpen()) {
		gui->setIO(true);
		serial.init9600();
	} else
		if (serial.getFd() > 0)
			serial.closePort();
	
	gui->setStatus(Gui::IDLE);
}
